# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import abc
import logging
import os
from pathlib import Path
import re
import shutil
from subprocess import Popen, PIPE
import yaml

from galaxy_importer import constants
from galaxy_importer import exceptions as exc
from galaxy_importer import schema
from galaxy_importer.utils import markup as markup_utils

default_logger = logging.getLogger(__name__)


class ContentLoader(metaclass=abc.ABCMeta):
    def __init__(self, content_type, rel_path, root, doc_strings=None, cfg=None, logger=None):
        """
        :param content_type: Content type.
        :param rel_path: Path to content file or dir, relative to root path.
        :param root: Collection root path.
        :param doc_strings: ansible-doc output for all plugins in collection
        :param logger: Optional logger instance.

        ==Example==
        Given:
            root='/tmp/tmpgjbj53c9/ansible_collections/my_namespace/nginx'
            rel_path='modules/plugins/storage/another_subdir/s3.py'
        Names will be:
            fq_collection_name: my_namespace.nginx
            name: s3
            path_name: storage.another_subdir.s3
            fq_name: my_namespace.nginx.storage.another_subdir.s3
        """
        self.content_type = content_type
        self.rel_path = rel_path
        self.root = root

        self.name = self._make_name(self.rel_path)
        self._validate_name()
        self.path_name = self._make_path_name(self.rel_path, self.name)

        self.doc_strings = doc_strings or {}
        self.cfg = cfg
        self.log = logger or default_logger

    @abc.abstractmethod
    def load(self):
        """Loads data from content inside collection.
        :return: Content object."""

    @staticmethod
    @abc.abstractmethod
    def _make_name(rel_path):
        """Returns content name generated from it's path."""

    @staticmethod
    @abc.abstractmethod
    def _make_path_name(rel_path, name):
        """Returns subdirectories as part content name.
        'sub1.sub2.mod' for plugins/modules/sub1/sub2/mod.py"""

    @staticmethod
    def _get_fq_collection_name(root):
        root_parts = Path(root).parts
        return "{}.{}".format(*root_parts[-2:])

    def _get_fq_name(self, root, path_name):
        return "{}.{}".format(
            self._get_fq_collection_name(root),
            path_name,
        )

    def _validate_name(self):
        if not re.match(constants.CONTENT_NAME_REGEXP, self.name):
            raise exc.ContentNameError(
                f"{self.content_type.value} name invalid format: {self.name}"
            )

    def _log_loading(self):
        self.log.info(f"Loading {self.content_type.value} {self.path_name}")


class PluginLoader(ContentLoader):
    def load(self):
        self._log_loading()
        doc_strings = self._get_plugin_doc_strings()

        if self.cfg.run_flake8:
            for line in self._run_flake8():
                self.log.warning(line)

        return schema.Content(
            name=self.path_name,
            content_type=self.content_type,
            doc_strings=doc_strings,
        )

    def _get_plugin_doc_strings(self):
        """Return plugin doc_strings, if exists, from collection doc_strings."""
        fq_name = self._get_fq_name(self.root, self.path_name)
        try:
            return self.doc_strings[self.content_type.value][fq_name]
        except KeyError:
            return None

    def _run_flake8(self):
        self.log.info(f"Linting {self.content_type.value} {self.path_name} via flake8...")

        if not shutil.which("flake8"):
            self.log.warning("flake8 not found, skipping")
            return

        cmd = [
            "flake8",
            "--exit-zero",
            "--isolated",
            "--extend-ignore",
            constants.FLAKE8_IGNORE_ERRORS,
            "--select",
            constants.FLAKE8_SELECT_ERRORS,
            "--max-line-length",
            str(constants.FLAKE8_MAX_LINE_LENGTH),
            "--",
            self.rel_path,
        ]

        self.log.debug("CMD: " + " ".join(cmd))
        proc = Popen(
            cmd,
            cwd=self.root,
            encoding="utf-8",
            stdout=PIPE,
        )

        for line in proc.stdout:
            yield line.strip()

    @staticmethod
    def _make_name(rel_path):
        return os.path.splitext(os.path.basename(rel_path))[0]

    @staticmethod
    def _make_path_name(rel_path, name):
        dirname_parts = Path(os.path.dirname(rel_path)).parts[2:]
        return ".".join(list(dirname_parts) + [name])


class ExtensionLoader(PluginLoader):
    @staticmethod
    def _make_path_name(rel_path, name):
        """Not expecting extensions to exist in subdirs like other content types,
        can simply return name
        """
        return name


class PlaybookLoader(ContentLoader):

    def load(self):
        self._log_loading()

        return schema.Content(
            name=self.path_name,
            content_type=self.content_type,
        )

    @staticmethod
    def _make_name(rel_path):
        return os.path.basename(rel_path)

    @staticmethod
    def _make_path_name(rel_path, name):
        dirname_parts = Path(os.path.dirname(rel_path)).parts[1:]
        return ".".join(list(dirname_parts) + [name])

    def _validate_name(self):
        return True


class RoleLoader(ContentLoader):
    def load(self):
        self._log_loading()
        description = self._get_metadata_description()
        readme = self._get_readme()

        return schema.Content(
            name=self.path_name,
            content_type=self.content_type,
            description=description,
            readme_file=readme.name,
            readme_html=markup_utils.get_html(readme),
        )

    @staticmethod
    def _make_name(rel_path):
        return os.path.basename(rel_path)

    @staticmethod
    def _make_path_name(rel_path, name):
        dirname_parts = Path(os.path.dirname(rel_path)).parts[1:]
        return ".".join(list(dirname_parts) + [name])

    def _get_readme(self):
        readme = markup_utils.get_readme_doc_file(os.path.join(self.root, self.rel_path))
        if not readme:
            raise exc.ContentLoadError("No role readme found.")
        return readme

    def _get_metadata_description(self):
        description = None
        meta_path = self._find_metadata_file_path(self.root, self.rel_path)

        if not meta_path:
            self.log.warning("Could not get role description, no role metadata found")
            return description

        with open(meta_path) as fp:
            try:
                role_metadata = yaml.safe_load(fp)
            except Exception:
                self.log.error("Error during parsing of role metadata")
        try:
            description = role_metadata["galaxy_info"]["description"]
        except KeyError:
            self.log.warning("No role description found in role metadata")
        return description

    @staticmethod
    def _find_metadata_file_path(root, rel_path):
        """Gets path to role metadata file."""
        for file in constants.ROLE_META_FILES:
            meta_path = os.path.join(root, rel_path, file)
            if os.path.exists(meta_path):
                return meta_path
        return None


def get_loader_cls(content_type):
    if content_type.category == constants.ContentCategory.PLAYBOOK:
        return PlaybookLoader
    elif content_type.category == constants.ContentCategory.ROLE:
        return RoleLoader
    elif content_type.category in [
        constants.ContentCategory.PLUGIN,
        constants.ContentCategory.MODULE,
    ]:
        return PluginLoader
    elif content_type.category == constants.ContentCategory.EXTENSION:
        return ExtensionLoader

    return None



from packaging.specifiers import SpecifierSet


def validate(requirement_string):
    """Validate requires_ansible specifier

    Args:
        requirement_string (str): The value of the 'requires_ansible' attribute in the runtime file.
    Raises:
        InvalidSpecifier: If the value of requirement_string is not a
            valid packaging.specifiers.SpecifierSet.

    Returns:
        bool: True if requirement_string is valid, otherwise should raise exception.
    """
    SpecifierSet(requirement_string)
    return True



from .collection import (
    CollectionLoader,
)

from .legacy_role import (
    LegacyRoleLoader,
)

from .content import (
    ContentLoader,
    PluginLoader,
    ExtensionLoader,
    RoleLoader,
    get_loader_cls,
)

from .doc_string import (
    DocStringLoader,
)

__all__ = (
    "CollectionLoader",
    "LegacyRoleLoader",
    "ContentLoader",
    "PluginLoader",
    "ExtensionLoader",
    "RoleLoader",
    "get_loader_cls",
    "DocStringLoader",
)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import json
import re

import attr
import semantic_version
import yaml

from galaxy_importer import config
from galaxy_importer import constants
from galaxy_importer import exceptions as exc
from galaxy_importer.utils.spdx_licenses import is_valid_license_id

MAX_LENGTH_AUTHOR = 64
MAX_LENGTH_LICENSE = 32
MAX_LENGTH_NAME = 64
MAX_LENGTH_NAMESPACE = 39
MAX_LENGTH_TAG = 64
MAX_LENGTH_URL = 2000
MAX_LENGTH_VERSION = 128

MAX_LEGACY_ROLE_LENGTH_VERSION = 10
MAX_LEGACY_ROLE_LENGTH_COMPANY = 50
MAX_LEGACY_ROLE_LENGTH_LICENSE = 50
MAX_LEGACY_ROLE_LENGTH_DESCRIPTION = 255

SHA1_LEN = 40
REQUIRED_TAG_LIST = [
    "application",
    "cloud",
    "database",
    "eda",
    "infrastructure",
    "linux",
    "monitoring",
    "networking",
    "security",
    "storage",
    "tools",
    "windows",
]


def convert_none_to_empty_dict(val):
    """Returns an empty dict if val is None."""

    # if val is not a dict or val 'None' return val
    # and let the validators raise errors later
    if val is None:
        return {}
    return val


_FILENAME_RE = re.compile(
    r"^(?P<namespace>\w+)-(?P<name>\w+)-(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$"
)


@attr.s(slots=True)
class CollectionFilename(object):
    namespace = attr.ib()
    name = attr.ib()
    version = attr.ib(converter=semantic_version.Version)

    def __str__(self):
        return f"{self.namespace}-{self.name}-{self.version}.tar.gz"

    @classmethod
    def parse(cls, filename):
        match = _FILENAME_RE.match(filename)
        if not match:
            raise ValueError("Invalid filename. Expected: {namespace}-{name}-{version}.tar.gz")

        return cls(**match.groupdict())

    @namespace.validator
    @name.validator
    def _validator(self, attribute, value):
        if not constants.NAME_REGEXP.match(value):
            raise ValueError("Invalid {0}: {1!r}".format(attribute.name, value))


@attr.s(frozen=True)
class CollectionInfo(object):
    """Represents collection_info metadata in collection manifest."""

    namespace = attr.ib(default=None)
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    license = attr.ib(factory=list)
    description = attr.ib(default=None)

    repository = attr.ib(default=None)
    documentation = attr.ib(default=None)
    homepage = attr.ib(default=None)
    issues = attr.ib(default=None)

    authors = attr.ib(factory=list)
    tags = attr.ib(factory=list)

    license_file = attr.ib(default=None)
    readme = attr.ib(default=None)

    dependencies = attr.ib(
        factory=dict,
        converter=convert_none_to_empty_dict,
        validator=attr.validators.instance_of(dict),
    )

    @property
    def label(self):
        return f"{self.namespace}.{self.name}"

    @staticmethod
    def value_error(msg):
        raise ValueError(f"Invalid collection metadata. {msg}") from None

    @namespace.validator
    @name.validator
    @version.validator
    @readme.validator
    @authors.validator
    @repository.validator
    def _check_required(self, attribute, value):
        """Check that value is present."""
        if not value:
            self.value_error(f"'{attribute.name}' is required")

    @namespace.validator
    @name.validator
    def _check_name(self, attribute, value):
        """Check value against name regular expression and max length."""
        if not re.match(constants.NAME_REGEXP, value):
            self.value_error(f"'{attribute.name}' has invalid format: {value}")
        if len(value) > MAX_LENGTH_NAME:
            self.value_error(
                f"'{attribute.name}' must not be greater than {MAX_LENGTH_NAME} characters"
            )

    @version.validator
    def _check_version_format(self, attribute, value):
        """Check that version is in semantic version format, and max length."""
        if not semantic_version.validate(value):
            self.value_error(
                f"Expecting 'version' to be in semantic version format, instead found '{value}'."
            )
        if len(value) > MAX_LENGTH_VERSION:
            self.value_error(f"'version' must not be greater than {MAX_LENGTH_VERSION} characters")

        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        if cfg.require_v1_or_greater:
            if semantic_version.Version(value) < semantic_version.Version("1.0.0"):
                self.value_error("Config is enabled that requires version to be 1.0.0 or greater.")

    @authors.validator
    @tags.validator
    @license.validator
    def _check_list_of_str(self, attribute, value):
        """Check that value is a list of strings."""
        err_msg = "Expecting '{attr}' to be a list of strings"
        if not isinstance(value, list):
            self.value_error(err_msg.format(attr=attribute.name))
        for list_item in value:
            if not isinstance(list_item, str):
                self.value_error(err_msg.format(attr=attribute.name))

    @license.validator
    def _check_licenses(self, attribute, value):
        """Check that all licenses in license list are valid."""
        for license in value:
            if len(license) > MAX_LENGTH_LICENSE:
                self.value_error(
                    f"Each license in 'licenses' list must not be greater than "
                    f"{MAX_LENGTH_LICENSE} characters"
                )
        invalid_licenses = [id for id in value if not is_valid_license_id(id)]
        if invalid_licenses:
            self.value_error(
                "Expecting 'license' to be a list of valid SPDX license "
                "identifiers, instead found invalid license identifiers: '{}' "
                "in 'license' value {}. "
                "For more info, visit https://spdx.org".format(", ".join(invalid_licenses), value)
            )

    @dependencies.validator
    def _check_dependencies_format(self, attribute, dependencies):
        """Check type and format of dependencies collection and version."""
        for collection, version_spec in dependencies.items():
            if not isinstance(collection, str):
                self.value_error("Expecting depencency to be string")
            if not isinstance(version_spec, str):
                self.value_error("Expecting depencency version to be string")

            try:
                namespace, name = collection.split(".")
            except ValueError:
                self.value_error(f"Invalid dependency format: '{collection}'")

            for value in [namespace, name]:
                if not re.match(constants.NAME_REGEXP, value):
                    self.value_error(
                        f"Invalid dependency format: '{value}' in '{namespace}.{name}'"
                    )

            if namespace == self.namespace and name == self.name:
                self.value_error("Cannot have self dependency")

            try:
                semantic_version.SimpleSpec(version_spec)
            except ValueError:
                self.value_error(
                    f"Dependency version spec range invalid: {collection} {version_spec}"
                )

    @tags.validator
    def _check_tags(self, attribute, value):
        """Check tags field for
        - presense of at least one required tag based on config
        - maximum count of tags
        - tag regular expression
        - tag max length
        """
        no_req_tag_err = f'At least one tag required from tag list: {", ".join(REQUIRED_TAG_LIST)}'

        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        if cfg.check_required_tags and not value:
            self.value_error(no_req_tag_err)

        if cfg.check_required_tags and (not any(tag in REQUIRED_TAG_LIST for tag in value)):
            self.value_error(no_req_tag_err)

        if not value:
            return

        if len(value) > constants.MAX_TAGS_COUNT:
            self.value_error(f"Expecting no more than {constants.MAX_TAGS_COUNT} tags in metadata")

        for tag in value:
            if not re.match(constants.NAME_REGEXP, tag):
                self.value_error(f"'tag' has invalid format: {tag}")
            if len(tag) > MAX_LENGTH_TAG:
                self.value_error(
                    f"Each tag in 'tags' list must not be greater than {MAX_LENGTH_TAG} characters"
                )

    @description.validator
    @repository.validator
    @documentation.validator
    @homepage.validator
    @issues.validator
    @license_file.validator
    @repository.validator
    def _check_non_null_str(self, attribute, value):
        """Check that if value is present, it must be a string."""
        if value is not None and not isinstance(value, str):
            self.value_error(f"'{attribute.name}' must be a string")

    @documentation.validator
    @homepage.validator
    @issues.validator
    @repository.validator
    def _check_max_length_url(self, attribute, value):
        """Check value is not greater than max length of url"""
        if not value:
            return
        if len(value) > MAX_LENGTH_URL:
            self.value_error(
                f"'{attribute.name}' must not be greater than {MAX_LENGTH_URL} characters"
            )

    @authors.validator
    def _check_max_length_author(self, attribute, value):
        """Check each value list item is not greater than max length of author"""
        for author in value:
            if len(author) > MAX_LENGTH_AUTHOR:
                self.value_error(
                    f"Each author in 'authors' list must not be greater than "
                    f"{MAX_LENGTH_AUTHOR} characters"
                )

    def __attrs_post_init__(self):
        """Checks called post init validation."""
        self._check_license_or_license_file()

    def _check_license_or_license_file(self):
        """Confirm mutually exclusive presence of license or license_file."""
        if bool(self.license) != bool(self.license_file):
            return

        if self.license and self.license_file:
            self.value_error("The 'license' and 'license_file' keys are mutually exclusive")

        self.value_error(
            "Valid values for 'license' or 'license_file' are required. "
            f"But 'license' ({self.license}) and 'license_file' ({self.license_file}) were invalid."
        )


@attr.s(frozen=True)
class CollectionArtifactFile(object):
    name = attr.ib()
    ftype = attr.ib()
    src_name = attr.ib(default=None)
    chksum_type = attr.ib(default="sha256")
    chksum_sha256 = attr.ib(default=None)
    format = attr.ib(default=1)


@attr.s(frozen=True)
class CollectionArtifactManifest(object):
    """Represents collection manifest metadata."""

    collection_info = attr.ib(type=CollectionInfo)
    format = attr.ib(default=1)

    file_manifest_file = attr.ib(
        default=None,
        type=CollectionArtifactFile,
        validator=attr.validators.optional(attr.validators.instance_of(CollectionArtifactFile)),
    )

    @classmethod
    def parse(cls, data):
        meta = json.loads(data)
        col_info = meta.pop("collection_info", None)
        meta["collection_info"] = CollectionInfo(**col_info)

        try:
            file_manifest_file = meta["file_manifest_file"]
        except KeyError as e:
            msg = "MANIFEST.json did not contain a 'file_manifest_file' item pointing to FILES.json"
            raise ValueError(msg) from e
        meta["file_manifest_file"] = CollectionArtifactFile(**file_manifest_file)
        return cls(**meta)


def convert_list_to_artifact_file_list(val):
    """Convert a list of dicts with file info into list of CollectionArtifactFile"""

    new_list = []
    for file_item in val:
        if isinstance(file_item, CollectionArtifactFile):
            new_list.append(file_item)
        else:
            artifact_file = CollectionArtifactFile(
                name=file_item["name"],
                ftype=file_item["ftype"],
                chksum_type=file_item.get("chksum_type"),
                chksum_sha256=file_item.get("chksum_sha256"),
            )
            new_list.append(artifact_file)

    return new_list


@attr.s(frozen=True)
class CollectionArtifactFileManifest(object):
    files = attr.ib(factory=list, converter=convert_list_to_artifact_file_list)

    format = attr.ib(default=1, validator=attr.validators.instance_of(int))

    @classmethod
    def parse(cls, data):
        artifact_file_manifest_data = json.loads(data)
        artifact_file_list = artifact_file_manifest_data["files"]
        return cls(files=artifact_file_list)


@attr.s(frozen=True)
class LegacyGalaxyInfo(object):
    """Represents legacy role metadata galaxy_info field."""

    role_name = attr.ib(default=None)
    namespace = attr.ib(default=None)
    author = attr.ib(default=None)
    description = attr.ib(default=None)
    company = attr.ib(default=None)
    issue_tracker_url = attr.ib(default=None)
    license = attr.ib(default=None)
    min_ansible_version = attr.ib(default=None)
    min_ansible_container_version = attr.ib(default=None)
    github_branch = attr.ib(default=None)
    platforms = attr.ib(factory=list)
    galaxy_tags = attr.ib(factory=list)

    @role_name.validator
    @namespace.validator
    @author.validator
    @description.validator
    @company.validator
    @issue_tracker_url.validator
    @license.validator
    @github_branch.validator
    def _validate_str(self, attribute, value):
        """Ensure value is of type str."""

        if value is not None and not isinstance(value, str):
            raise exc.LegacyRoleSchemaError(f"{attribute.name} must be a string")

    @platforms.validator
    def _validate_list_dict(self, attribute, value):
        """Ensure value is of type list[dict]."""

        if value is not None:
            if not isinstance(value, list):
                raise exc.LegacyRoleSchemaError(f"{attribute.name} must be a list")
            if not all(isinstance(element, dict) for element in value):
                raise exc.LegacyRoleSchemaError(f"{attribute.name} must be a list of dictionaries")

    @galaxy_tags.validator
    def _validate_list_str(self, attribute, value):
        """Ensure value is of type list[str]."""

        if value is not None:
            if not isinstance(value, list):
                raise exc.LegacyRoleSchemaError(f"{attribute.name} must be a list")
            if not all(isinstance(element, str) for element in value):
                raise exc.LegacyRoleSchemaError(f"{attribute.name} must be a list of strings")

    @role_name.validator
    def _validate_role_name(self, attribute, value):
        """Ensure role name matches the regular expression."""

        if value is not None and not constants.LEGACY_ROLE_NAME_REGEXP.match(value):
            raise exc.LegacyRoleSchemaError(f"role name {value} is invalid")

    @namespace.validator
    def _validate_namespace(self, attribute, value):
        """Ensure namespace matches the regular expression."""

        if value is not None and (
            not constants.LEGACY_NAMESPACE_REGEXP.match(value) or len(value) > MAX_LENGTH_NAMESPACE
        ):
            raise exc.LegacyRoleSchemaError(f"namespace {value} is invalid")

    @author.validator
    def _validate_author(self, attribute, value):
        """Ensure the author value is not too long."""

        if value is not None and len(value) > MAX_LENGTH_AUTHOR:
            raise exc.LegacyRoleSchemaError(
                f"author must not exceed {MAX_LENGTH_AUTHOR} characters"
            )

    @description.validator
    def _validate_description(self, attribute, value):
        if value is not None and len(value) > MAX_LEGACY_ROLE_LENGTH_DESCRIPTION:
            raise exc.LegacyRoleSchemaError(
                f"description must not exceed {MAX_LEGACY_ROLE_LENGTH_DESCRIPTION} characters"
            )

    @company.validator
    def _validate_company(self, attribute, value):
        if value is not None and len(value) > MAX_LEGACY_ROLE_LENGTH_COMPANY:
            raise exc.LegacyRoleSchemaError(
                f"company must not exceed {MAX_LEGACY_ROLE_LENGTH_DESCRIPTION} characters"
            )

    @issue_tracker_url.validator
    def _validate_url(self, attribute, value):
        """Ensure a URL is not too long."""

        if value is not None and len(value) > MAX_LENGTH_URL:
            raise exc.LegacyRoleSchemaError(f"url must not exceed {MAX_LENGTH_URL} characters")

    @license.validator
    def _validate_license(self, attribute, value):
        """Ensure the license is not too long."""

        if value is not None and len(value) > MAX_LEGACY_ROLE_LENGTH_LICENSE:
            raise exc.LegacyRoleSchemaError(
                f"role license must not exceed {MAX_LEGACY_ROLE_LENGTH_LICENSE} characters"
            )

    @min_ansible_version.validator
    @min_ansible_container_version.validator
    def _validate_version(self, attribute, value):
        """Ensure a version is not too long."""

        if value is not None and len(str(value)) > MAX_LEGACY_ROLE_LENGTH_VERSION:
            raise exc.LegacyRoleSchemaError(
                f"version for {attribute.name} must not exceed"
                f"{MAX_LEGACY_ROLE_LENGTH_VERSION} characters"
            )

    @galaxy_tags.validator
    def _validate_tags(self, attribute, value):
        """Ensure tags are not too long."""

        if value is not None:
            if any(len(element) > MAX_LENGTH_TAG for element in value):
                raise exc.LegacyRoleSchemaError(f"tag must not exceed {MAX_LENGTH_TAG} characters")


@attr.s(frozen=True)
class LegacyMetadata:
    """Represents legacy role metadata."""

    galaxy_info = attr.ib(default=None, type=LegacyGalaxyInfo)
    dependencies = attr.ib(factory=list)

    @classmethod
    def parse(cls, data):
        with open(data, "r") as fh:
            metadata = yaml.safe_load(fh)
            if not isinstance(metadata, dict):
                raise exc.LegacyRoleSchemaError("metadata must be in the form of a yaml dictionary")
            if "galaxy_info" not in metadata:
                raise exc.LegacyRoleSchemaError("galaxy_info field not found in metadata")
            if not isinstance(metadata["galaxy_info"], dict):
                raise exc.LegacyRoleSchemaError("galaxy_info field must contain a dictionary")
            try:
                galaxy_info = LegacyGalaxyInfo(**metadata["galaxy_info"])
            except TypeError as e:
                raise exc.LegacyRoleSchemaError(f"unknown field in galaxy_info: {e}") from e
            dependencies = metadata.get("dependencies", list())
        return cls(galaxy_info, dependencies)

    @dependencies.validator
    def _validate_dependencies(self, attribute, value):
        if not isinstance(value, list):
            raise exc.LegacyRoleSchemaError(
                "dependencies must be either a list of strings or a list of dictionaries."
            )
        for dependency in value:
            if not isinstance(dependency, str) and not isinstance(dependency, dict):
                raise exc.LegacyRoleSchemaError(
                    "dependencies must be either a list of strings or a list of dictionaries."
                )

            _dependency = dependency
            if isinstance(dependency, dict):
                if "role" in dependency:
                    _dependency = dependency.get("role", "")
                elif "name" in dependency:
                    _dependency = dependency.get("name", "")
                elif "src" in dependency:
                    _dependency = dependency.get("src", "")
                else:
                    raise exc.LegacyRoleSchemaError(
                        "dependency must include either the 'role,' 'name,' or 'src' keyword."
                    )

            if _dependency.count(".") != 1:
                raise exc.LegacyRoleSchemaError(
                    f"{dependency} must have namespace and name separated by '.'"
                )


@attr.s(frozen=True)
class ResultContentItem(object):
    name = attr.ib()
    content_type = attr.ib()
    description = attr.ib()


@attr.s(frozen=True)
class ImportResult(object):
    """Result of the import process, collection metadata, and contents."""

    metadata = attr.ib(default=None, type=CollectionInfo)
    docs_blob = attr.ib(factory=dict)
    contents = attr.ib(factory=list, type=ResultContentItem)
    custom_license = attr.ib(default=None)
    requires_ansible = attr.ib(default=None)


@attr.s(frozen=True)
class LegacyImportResult(object):
    """Result of legacy import with namespace, name, metadata, and readme."""

    namespace = attr.ib()
    name = attr.ib()
    metadata = attr.ib(type=LegacyMetadata)
    readme_file = attr.ib(default=None)
    readme_html = attr.ib(default=None)

    @name.validator
    def _validate_name(self, attribute, value):
        if constants.LEGACY_ROLE_NAME_REGEXP.match(value) is None:
            raise exc.LegacyRoleSchemaError(f"role name {value} is invalid")

    @metadata.validator
    def _ensure_no_self_dependency(self, attribute, value):
        for dependency in self.metadata.dependencies:
            _dependency = dependency
            if isinstance(dependency, dict):
                if "role" in dependency:
                    _dependency = dependency.get("role", "")
                elif "name" in dependency:
                    _dependency = dependency.get("name", "")
                elif "src" in dependency:
                    _dependency = dependency.get("src", "")

            namespace, name = _dependency.split(".")
            if self.namespace == namespace and self.name == name:
                raise exc.LegacyRoleSchemaError(
                    f"role {self.namespace}.{self.name} cannot depend on itself"
                )


@attr.s
class Content(object):
    """Represents content found in a collection."""

    name = attr.ib()
    content_type = attr.ib(type=constants.ContentType)
    doc_strings = attr.ib(factory=dict)
    description = attr.ib(default=None)
    readme_file = attr.ib(default=None)
    readme_html = attr.ib(default=None)

    def __attrs_post_init__(self):
        """Set description if a plugin has doc_strings populated."""
        if not self.doc_strings:
            return
        if not self.doc_strings.get("doc", None):
            return
        self.description = self.doc_strings["doc"].get("short_description", None)


@attr.s(frozen=True)
class RenderedDocFile(object):
    """Name and html of a documenation file, part of DocsBlob."""

    name = attr.ib(default=None)
    html = attr.ib(default=None)


@attr.s(frozen=True)
class DocsBlobContentItem(object):
    """Documenation for piece of content, part of DocsBlob."""

    content_name = attr.ib()
    content_type = attr.ib()
    doc_strings = attr.ib(factory=dict)
    readme_file = attr.ib(default=None)
    readme_html = attr.ib(default=None)


@attr.s(frozen=True)
class DocsBlob(object):
    """All documenation that is part of a collection."""

    collection_readme = attr.ib(type=RenderedDocFile)
    documentation_files = attr.ib(factory=list, type=RenderedDocFile)
    contents = attr.ib(factory=list, type=DocsBlobContentItem)



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.


import os
import shutil
import subprocess

from galaxy_importer import config
from galaxy_importer.ansible_test.runners.base import BaseTestRunner

config_data = config.ConfigFile.load()
cfg = config.Config(config_data=config_data)


class LocalAnsibleTestRunner(BaseTestRunner):
    """Run ansible-test locally with --docker or using venv."""

    def run(self):
        if not shutil.which("ansible"):
            self.log.error("ansible not found, skipping ansible-test")
            return

        version_proc = subprocess.Popen(
            [
                "/usr/bin/env",
                f"ANSIBLE_LOCAL_TEMP={cfg.ansible_local_tmp}",
                "ansible",
                "--version",
            ],
            stdout=subprocess.PIPE,
            encoding="utf-8",
        )
        self.log.info(f"Using {list(version_proc.stdout)[0].rstrip()}")

        suffix = f"ansible_collections/{self.metadata.namespace}/{self.metadata.name}/"
        collection_dir = os.path.join(self.dir, suffix)

        cmd = [
            "/usr/bin/env",
            f"ANSIBLE_LOCAL_TEMP={cfg.ansible_local_tmp}",
            "ansible-test",
            "sanity",
            "--docker",
            "--color",
            "yes",
            "--failure-ok",
        ]

        collection_name = f"{self.metadata.namespace}-{self.metadata.name}-{self.metadata.version}"
        self.log.info(f"Running ansible-test sanity on {collection_name} ...")
        self.log.info(f'{" ".join(cmd)}')

        proc = subprocess.Popen(
            cmd,
            cwd=collection_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
        )

        for line in proc.stdout:
            self.log.info(line.strip())

        return_code = proc.wait()
        if return_code != 0:
            self.log.error(
                "An exception occurred in {}, returncode={}, collection={}".format(
                    " ".join(cmd), return_code, collection_name
                )
            )



import logging
import os

from galaxy_importer import exceptions as exc
from galaxy_importer import __version__
from galaxy_importer.utils import markup

default_logger = logging.getLogger(__name__)


def convert_markdown(dirname, logger=None):
    """Convert README.md to html.

    NOTE: README.md must exist in dirname

    :param dirname: directory of role.
    :param logger: Optional logger instance.

    :raises exc.ImporterError: On errors that fail the import process.

    :return: html markup of supplied README.md file
    """
    logger = logger or default_logger
    logger.info(f"Converting markdown with galaxy-importer {__version__}")

    # Ensure that dirname exists.
    if not os.path.exists(dirname):
        raise exc.ImporterError(f"Path does not exist: {dirname}")

    return _convert_markdown(dirname, logger)


def _convert_markdown(dirname, logger):
    doc_file = markup.get_readme_doc_file(dirname)
    if not doc_file:
        raise exc.ImporterError(f"Path does not contain README.md: {dirname}")
    else:
        logger.info(f"Processing {dirname}{doc_file.name}")
    return {"html": markup.get_html(doc_file)}



"""Utils for manipulating strings"""


def removeprefix(target, prefix):
    """Remove a prefix from a string, based on 3.9 str.removeprefix()"""
    if target.startswith(prefix):
        return target[len(prefix) :]
    else:
        return target[:]



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import abc
import logging

default_logger = logging.getLogger(__name__)


class BaseTestRunner(object):
    """
    :param dir: Dir where collection is extracted, used by local runner.
    :param metadata: Collection metadata, used by local runner.
    :param filepath: Path where archive file is located.
    """

    def __init__(self, dir="", metadata="", file=None, filepath=None, logger=None, file_url=None):
        self.log = logger or default_logger
        self.dir = dir
        self.filepath = filepath
        self.metadata = metadata
        self.file = file
        self.file_url = file_url

    @abc.abstractmethod
    def run(self):
        pass



import json
import logging

from pkg_resources import resource_stream


log = logging.getLogger(__name__)

_SPDX_LICENSES = None
_SPDX_LICENSES_FILE = "spdx_licenses.json"


def _load_spdx():
    try:
        with resource_stream(__name__, _SPDX_LICENSES_FILE) as stream:
            return json.load(stream)
    except EnvironmentError as exc:
        log.warning(
            "Unable to open %s to load the list of acceptable open source licenses: %s",
            _SPDX_LICENSES_FILE,
            exc,
        )
        log.exception(exc)
        return {}


def _get_spdx():
    """Gets the spdx_license info, so it is only loaded from disk once."""
    global _SPDX_LICENSES

    if not _SPDX_LICENSES:
        _SPDX_LICENSES = _load_spdx()

    return _SPDX_LICENSES


def is_valid_license_id(license_id):
    """Check if license_id is valid and non-deprecated SPDX ID."""
    if license_id is None:
        return False

    valid_license_ids = _get_spdx()
    valid = valid_license_ids.get(license_id, None)
    if valid is None:
        return False

    # license was in list, but is deprecated
    if valid and valid.get("deprecated", None):
        return False

    return True



# (c) 2012-2023, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging
import os
import yaml

from galaxy_importer import constants
from galaxy_importer import exceptions as exc
from galaxy_importer.utils import requires_ansible_version

default_logger = logging.getLogger(__name__)


class RuntimeFileParser:
    """Load meta/runtime.yml."""

    def __init__(self, collection_path):
        self.path = os.path.join(collection_path, "meta/runtime.yml")
        self.data = None
        self._load()

    def _load(self):
        if not os.path.exists(self.path):
            return
        with open(self.path) as fp:
            try:
                self.data = yaml.safe_load(fp)
            except Exception:
                raise exc.FileParserError("Error during parsing of runtime.yml")

    def get_requires_ansible(self):
        if not self.data:
            raise exc.FileParserError(
                "'requires_ansible' in meta/runtime.yml is mandatory, "
                "but no meta/runtime.yml found"
            )
        requires_ansible = self.data.get("requires_ansible")
        if not requires_ansible:
            raise exc.FileParserError(
                "'requires_ansible' in meta/runtime.yml is mandatory, "
                "but 'requires_ansible' not found"
            )
        if len(requires_ansible) > constants.MAX_LENGTH_REQUIRES_ANSIBLE:
            raise exc.FileParserError(
                f"'requires_ansible' must not be greater than "
                f"{constants.MAX_LENGTH_REQUIRES_ANSIBLE} characters"
            )
        try:
            requires_ansible_version.validate(requires_ansible)
            return requires_ansible
        except ValueError:
            raise exc.FileParserError("'requires_ansible' is not a valid requirement specification")


class ExtensionsFileParser:
    """Load meta/extensions.yml."""

    def __init__(self, collection_path):
        self.path = os.path.join(collection_path, "meta/extensions.yml")
        self.data = None
        self._load()

    def _load(self):
        if not os.path.exists(self.path):
            return
        with open(self.path) as fp:
            try:
                self.data = yaml.safe_load(fp)
            except Exception:
                raise exc.FileParserError("Error during parsing of extensions.yml")

    def get_extension_dirs(self):
        try:
            if not self.data:
                return []
            # TODO: consider check that ext_dir actually exists and fail import if not
            # TODO: consider check that ext_dir has exactly 3 dirs otherwise fail
            return [ext["args"]["ext_dir"] for ext in self.data["extensions"]]
        except KeyError:
            raise exc.FileParserError("'meta/extensions.yml is not in the expected format'")



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import shutil
from subprocess import Popen, PIPE, STDOUT

from galaxy_importer import config
from galaxy_importer import exceptions
from galaxy_importer.ansible_test.builders.local_image_build import Build
from galaxy_importer.ansible_test.runners.base import BaseTestRunner


class LocalImageTestRunner(BaseTestRunner):
    """Run image locally with docker or podman."""

    def run(self):
        cfg = config.Config(config_data=config.ConfigFile.load())

        build = Build(
            self.filepath,
            f"{self.metadata.namespace}-{self.metadata.name}-{self.metadata.version}",
            cfg,
            self.log,
        )

        container_engine = build.get_container_engine(cfg)

        if not shutil.which(container_engine):
            self.log.warning(f'"{container_engine}" not found, skipping ansible-test sanity')
            return

        image_id = build.build_image()

        self.log.info("Running image...")
        self._run_image(image_id=image_id, container_engine=container_engine)

        build.cleanup()

    def _run_image(self, image_id, container_engine):
        cmd = [container_engine, "run", image_id, "LOCAL_IMAGE_RUNNER"]
        proc = Popen(
            cmd,
            stdout=PIPE,
            stderr=STDOUT,
            encoding="utf-8",
        )

        for line in proc.stdout:
            self.log.info(line.strip())

        return_code = proc.wait()
        if return_code != 0:
            raise exceptions.AnsibleTestError(
                "An exception occurred in {}, returncode={}".format(" ".join(cmd), return_code)
            )



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging

from .local_ansible_test import LocalAnsibleTestRunner
from .local_image import LocalImageTestRunner
from .openshift_job import OpenshiftJobTestRunner


default_logger = logging.getLogger(__name__)


def get_runner(cfg):
    """Decide which runner class to run ansible-test based on config."""

    if not cfg.run_ansible_test:
        return None

    if cfg.infra_osd:
        return OpenshiftJobTestRunner

    if cfg.ansible_test_local_image:
        return LocalImageTestRunner

    return LocalAnsibleTestRunner



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import configparser
import os


FILENAME = "galaxy-importer.cfg"
FILE_LOCATIONS = [
    f"/etc/galaxy-importer/{FILENAME}",
    os.path.join(os.path.dirname(__file__), FILENAME),
]
IMPORTER_INI_SECTION = "galaxy-importer"


def parse_string_to_bool(val):
    if val is None:
        return val
    if val.lower() in ["yes", "1", "true"]:
        return True
    if val.lower() in ["no", "0", "false"]:
        return False
    return val


class Config(object):
    """Configuration for galaxy-importer."""

    DEFAULTS = {
        "ansible_local_tmp": "~/.ansible/tmp",
        "ansible_test_local_image": False,
        "check_changelog": True,
        "check_required_tags": False,
        "infra_osd": False,
        "local_image_docker": False,
        "log_level_main": "INFO",
        "require_v1_or_greater": False,
        "run_ansible_doc": True,
        "run_ansible_lint": True,
        "offline_ansible_lint": True,
        "run_ansible_test": False,
        "run_flake8": False,
        "tmp_root_dir": None,
    }

    def __init__(self, config_data=None):
        """Set config values to default, updated with any passed config_data."""
        _data = {}
        _data.update(self.DEFAULTS)
        _data.update(config_data or {})
        self.__dict__.update(_data)

        # Allow environment overrides for testing
        for key in self.__dict__.keys():
            env_key = "GALAXY_IMPORTER_" + key.upper()
            if env_key in os.environ:
                self.__dict__[key] = parse_string_to_bool(os.environ[env_key])


class ConfigFile(object):
    """Load config from file and return dictionary."""

    @staticmethod
    def load():
        env_config = os.getenv("GALAXY_IMPORTER_CONFIG")
        if env_config:
            FILE_LOCATIONS.insert(0, env_config)
        config_parser_data = ConfigFile._load_file(FILE_LOCATIONS)
        return ConfigFile._to_dictionary(config_parser_data)

    @staticmethod
    def _load_file(file_locations):
        file_path = None
        for f in file_locations:
            if os.path.isfile(f):
                file_path = f
                break

        if file_path:
            config_parser = configparser.ConfigParser()
            config_parser.read(file_path)
            if IMPORTER_INI_SECTION not in config_parser:
                return {}
            return config_parser[IMPORTER_INI_SECTION]
        return {}

    @staticmethod
    def _to_dictionary(config_parser_data):
        """Turn from configparser object in to dictionary, with booleans."""
        config_data = {}
        for key in list(config_parser_data):
            try:
                config_data[key] = config_parser_data.getboolean(key)
            except ValueError:
                config_data[key] = config_parser_data.get(key)
        return config_data



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

from copy import deepcopy
import json
import logging
import os
import shutil
from subprocess import Popen, PIPE

from galaxy_importer import constants

default_logger = logging.getLogger(__name__)


class DocStringLoader:
    """Process ansible-doc doc strings for entire collection.

    Load by calling ansible-doc once in batch for each plugin type."""

    def __init__(self, path, fq_collection_name, cfg, logger=None):
        self.path = path
        self.fq_collection_name = fq_collection_name
        self.cfg = cfg
        self.log = logger or default_logger

    def load(self):
        self.log.info("Getting doc strings via ansible-doc")
        docs = {}

        if not shutil.which("ansible-doc"):
            self.log.warning("ansible-doc not found, skipping loading of docstrings")
            return docs

        for plugin_type in constants.ANSIBLE_DOC_SUPPORTED_TYPES:
            # use ansible-doc to list all the plugins of this type
            found_plugins = self._run_ansible_doc_list(plugin_type)
            plugins = sorted(list(found_plugins.keys()))

            if not plugins:
                continue

            data = self._run_ansible_doc(plugin_type, plugins)
            data = self._process_doc_strings(data)
            docs[plugin_type] = data

        return docs

    @property
    def _collections_path(self):
        return "/".join(self.path.split("/")[:-3])

    @property
    def _base_ansible_doc_cmd(self):
        return [
            "/usr/bin/env",
            f"ANSIBLE_COLLECTIONS_PATHS={self._collections_path}",
            f"ANSIBLE_COLLECTIONS_PATH={self._collections_path}",
            f"ANSIBLE_LOCAL_TEMP={self.cfg.ansible_local_tmp}",
            "ansible-doc",
        ]

    def _get_plugins(self, plugin_dir):
        """Get list of fully qualified plugin names inside directory.

        Ex: ['google.gcp.service_facts', 'google.gcp.storage.subdir2.gc_storage']
        """
        plugins = []
        for root, _, files in os.walk(plugin_dir):
            for filename in files:
                if not filename.endswith(".py") or filename == "__init__.py":
                    continue
                file_path = os.path.join(root, filename)
                sub_dirs = os.path.relpath(root, plugin_dir)

                fq_name_parts = [self.fq_collection_name]
                if sub_dirs and sub_dirs != ".":
                    fq_name_parts.extend(sub_dirs.split("/"))
                fq_name_parts.append(os.path.basename(file_path)[:-3])

                plugins.append(".".join(fq_name_parts))

        return plugins

    def _run_ansible_doc_list(self, plugin_type):
        """Use ansible-doc to get a list of plugins for the collection by type."""
        cmd = self._base_ansible_doc_cmd + [
            "--list",
            "--type",
            plugin_type,
            "--json",
            self.fq_collection_name,
        ]
        self.log.debug("CMD: {}".format(" ".join(cmd)))
        proc = Popen(cmd, cwd=self._collections_path, stdout=PIPE, stderr=PIPE)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            self.log.error(
                'Error running ansible-doc: cmd="{cmd}" returncode="{rc}" {err}'.format(
                    cmd=" ".join(cmd), rc=proc.returncode, err=stderr
                )
            )
            return {}
        return json.loads(stdout)

    def _run_ansible_doc(self, plugin_type, plugins):
        cmd = (
            self._base_ansible_doc_cmd
            + [
                "--type",
                plugin_type,
                "--json",
            ]
            + plugins
        )
        self.log.debug("CMD: {}".format(" ".join(cmd)))
        proc = Popen(cmd, cwd=self._collections_path, stdout=PIPE, stderr=PIPE)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            self.log.error(
                'Error running ansible-doc: cmd="{cmd}" returncode="{rc}" {err}'.format(
                    cmd=" ".join(cmd), rc=proc.returncode, err=stderr
                )
            )
            return {}
        return json.loads(stdout)

    def _process_doc_strings(self, doc_strings):
        processed_doc_strings = {}
        for plugin_key, value in doc_strings.items():
            processed_doc_strings[plugin_key] = self._transform_doc_strings(value, self.log)
        return processed_doc_strings

    @staticmethod
    def _transform_doc_strings(data, logger=default_logger):
        """Transform data meant for UI tables into format suitable for UI."""

        def dict_to_named_list(dict_of_dict):
            """Return new list of dicts for given dict of dicts."""
            try:
                return [{"name": key, **deepcopy(dict_of_dict[key])} for key in dict_of_dict.keys()]
            except TypeError:
                logger.warning(f"Expected this to be a dictionary of dictionaries: {dict_of_dict}")
                return [
                    {"name": key, **deepcopy(dict_of_dict[key])}
                    for key in dict_of_dict.keys()
                    if isinstance(key, dict)
                ]

        def handle_nested_tables(obj, table_key):
            """Recurse over dict to replace nested tables with updated format."""
            if table_key in obj.keys() and isinstance(obj[table_key], dict):
                obj[table_key] = dict_to_named_list(obj[table_key])
                for row in obj[table_key]:
                    handle_nested_tables(row, table_key)

        doc = data.get("doc", {})
        if isinstance(doc.get("options"), dict):
            doc["options"] = dict_to_named_list(doc["options"])
            for d in doc["options"]:
                handle_nested_tables(d, table_key="suboptions")

        ret = data.get("return", None)
        if ret and isinstance(ret, dict):
            data["return"] = dict_to_named_list(ret)
            for d in data["return"]:
                handle_nested_tables(d, table_key="contains")

        return data



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

from collections import namedtuple
import logging
import os
import shutil
import subprocess
import tarfile
import tempfile

import attr

from galaxy_importer import config
from galaxy_importer import exceptions as exc
from galaxy_importer.loaders import CollectionLoader
from galaxy_importer.ansible_test import runners
from galaxy_importer import __version__

default_logger = logging.getLogger(__name__)

CollectionFilename = namedtuple("CollectionFilename", ["namespace", "name", "version"])


def import_collection(
    file=None,
    filename=None,
    file_url=None,
    git_clone_path=None,
    output_path=None,
    logger=None,
    cfg=None,
):
    """Process import on collection artifact file object.

    :param file: file handle of collection artifact.
    :param filename: namedtuple of CollectionFilename.
    :param file_url: storage url of collection artifact.
    :param git_clone_path: path to git repo directory of collection pre artifact build.
    :param output_path: path where collection build tarball file will be written.
    :param logger: Optional logger instance.
    :param cfg: Optional config.

    :raises exc.ImporterError: On errors that fail the import process.

    :return: metadata if `file`  provided, (metadata, filepath) if `git_clone_path` provided
    """

    logger = logger or default_logger
    logger.info(f"Importing with galaxy-importer {__version__}")
    if not cfg:
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)

    if (file and git_clone_path) or not (file or git_clone_path):
        raise exc.ImporterError("Expected either 'file' or 'git_clone_path' to be populated")

    if git_clone_path:
        filepath = _build_collection(git_clone_path, output_path, logger)
        with open(filepath, "rb") as fh:
            metadata = _import_collection(fh, filename=None, file_url=None, logger=logger, cfg=cfg)
        return (metadata, filepath)

    return _import_collection(file, filename, file_url, logger, cfg)


def sync_collection(git_clone_path, output_path, logger=None, cfg=None):
    """Process collection metadata without linting to support pulp-ansible sync.

    Call _import_collection() with an overridden config to
    process metadata without linting and without running ansible-test.
    """

    logger = logger or default_logger
    if not cfg:
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)

    cfg.run_ansible_test = False
    cfg.run_ansible_lint = False
    cfg.run_flake8 = False

    filepath = _build_collection(git_clone_path, output_path, logger)
    with open(filepath, "rb") as fh:
        metadata = _import_collection(fh, filename=None, file_url=None, logger=logger, cfg=cfg)
    return (metadata, filepath)


def _build_collection(git_clone_path, output_path, logger=None):
    """Runs `ansible-galaxy collection build` and returns artifact filepath."""

    logger = logger or default_logger
    logger.info("Building collection tarball with ansible-galaxy collection build")

    cmd = ["ansible-galaxy", "collection", "build", "--output-path", output_path]
    result = subprocess.run(cmd, cwd=git_clone_path, capture_output=True)

    if result.returncode != 0:
        raise exc.ImporterError(
            "Error running `ansible-galaxy collection build`: {}".format(
                result.stderr.decode("utf-8").rstrip()
            )
        )

    # TODO: use regex to get filename from stdout, compine with output_path in case cli output
    # ever changes from: Created collection for <namespace>.<name> at /<path>/<artifact>.tar.gz
    stdout = result.stdout.decode("utf-8").rstrip()
    filepath = stdout.split(" ")[-1]
    return filepath


def _import_collection(file, filename, file_url, logger, cfg):
    """Returns collection version metadata."""

    with tempfile.TemporaryDirectory(dir=cfg.tmp_root_dir) as tmp_dir:
        sub_path = "ansible_collections/placeholder_namespace/placeholder_name"
        extract_dir = os.path.join(tmp_dir, sub_path)
        os.makedirs(extract_dir)
        _extract_archive(fileobj=file, extract_dir=extract_dir)

        data = CollectionLoader(extract_dir, filename, cfg=cfg, logger=logger).load()
        logger.info("Collection loading complete")

        ansible_test_runner = runners.get_runner(cfg=cfg)
        if ansible_test_runner:
            filepath = file.name
            if not os.path.exists(filepath):
                filepath = os.path.join(tmp_dir, "archive.tar.gz")
                file.seek(0)
                with open(filepath, "wb") as newfile:
                    shutil.copyfileobj(file, newfile)

            file.seek(0)
            ansible_test_runner(
                dir=tmp_dir,
                metadata=data.metadata,
                file=file,
                filepath=filepath,
                file_url=file_url,
                logger=logger,
            ).run()

    return attr.asdict(data)


def _extract_archive(fileobj, extract_dir):
    fileobj.seek(0)
    with tarfile.open(fileobj=fileobj, mode="r") as tf:
        for item in tf.getmembers():
            if item.name.startswith("/") or "../" in item.name:
                raise exc.ImporterError("Invalid file paths detected.")
            if item.linkname:
                # Ensure the link target is within the extraction root
                link_target = os.path.normpath(
                    os.path.join(extract_dir, os.path.dirname(item.name), item.linkname)
                )
                if not link_target.startswith(os.path.abspath(extract_dir)):
                    raise exc.ImporterError("Invalid link target detected.")
        tf.extractall(extract_dir)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.
